/*____________________________________________________________________________*\
 *

 Copyright (c) 1997-2003 John Roy, Holger Zimmermann. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. The name of the author may not be used to endorse or promote products
    derived from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/HTTP/ExprLog.cpp,v $
 * $Date: 2003/05/13 18:42:03 $
 *
 Description:
		- Document
\*____________________________________________________________________________*/
//$SourceTop:$

#include <stdio.h>
#include <fstream.h>
#include <ctype.h>

#include "HandBase.h"
#include "HTTPCore.h"
#include "PIStrStr.h"
#include "Pi3Expr.h"

/*____________________________________________________________________________*\
 *
 Description:
\*____________________________________________________________________________*/
#define KEY_CONF_FILE			"File"
#define KEY_CONF_OPENMODE		"OpenMode"
#define KEY_CONF_EXPRESSION		"Expression"

#if 0
	/*
	** HTML documentation for this handler
	*/
/*___+++HTMLDOC_BEGIN+++___*/
Name:
	ExpressionLogger

Description:
	Writes a logfile with flexible data and format.

Options:
<H5>Overview</H5>
<TABLE BORDER=1>
<TH>Option
<TH>Default
<TH>Values
<TH>Short Description
<TH>Example(s)

<TR>
<TD>File
<TD>+
<TD>&lt;filepath&gt;
<TD>Relative or Absolute path to log file
<TD>File="Logs/access.txt"

<TR>
<TD>Expression
<TD>+
<TD>&lt;Pi3Expression&gt;
<TD>Pi3Expression to write 
<TD>Expression="$A $h - [$t] $r $s $b"

<TR>
<TD>OpenMode
<TD>w
<TD>w or a
<TD>File mode to open logfile with
<TD>OpenMode="a"

</TABLE>
<STRONG>-</STRONG> in the <IT>default</IT> indicates no default<BR>
<STRONG>+</STRONG> in the <IT>default</IT> indicates the field is mandatory<BR>

<H4>Description of Options</H4>
<H5>
	File
</H5>
Relative or absolute path to logfile.
<H5>
	Expression
</H5>
Pi3Expression to write to logfile every time this handler is invoked
<H5>
	OpenMode
</H5>
Specifies the way the log file will be opened. 'A' is append, the file
is opened so as to appended to existing data. 'W' the file is
truncated on open if it already exists. In both cases the file is 
created if it does not already exist.

Phase:
	LOG

Returns:
	PIAPI_ERROR if this handler was invoked for a phase other than
	LOG. Otherwise PIAPI_CONTINUE;

Example:
	<PRE>
	&lt;Object&gt;
		Name ExpressionLogger
		Class ExpressionLoggerClass
		OpenMode "a"
		Expression "$A $h - [$t] $r $s $b"
	&lt;/Object&gt;

	&lt;Object&gt;
		...
		Log ExpressionLogger File="Logs/access.txt"
		...
	&lt;/Object&gt;

	</PRE>
-->

/*___+++HTMLDOC_END+++___*/
#endif

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class ExpressionLogger : public HandlerBaseHTTP
{
private:
	/* ---
	Configuration data
	--- */
	PIString sFilePath;					/* path to file to open */
	PIString sOpenMode;					/* file open mode */
	Pi3Expression *pExpr;				/* expression to log */
	PIPLATFORM_FD tFd;					/* file descriptor */

protected:
	int Parameter( const char *pVariable, const char *pValue,
		const char *pWhere )
		{
		assert( pVariable && pValue );
		PIOStrStream osErr;
		osErr << pWhere << "ExpressionLogger: ";
		if ( !PIUtil_stricmp( KEY_CONF_OPENMODE, pVariable ) )
			{
			sOpenMode = pValue;
			}
		else if ( !PIUtil_stricmp( KEY_CONF_EXPRESSION, pVariable ) )
			{
			if ( pExpr )
				{
				osErr << "'Expression' may only be specified once." << ends;
				CONFIG_ERR( Object(), osErr.str() );
				return 0;
				};
			Pi3String *pError = Pi3String_new( 0 );
			pExpr = Pi3Expression_new( pValue, 0, pError );
			if ( !pExpr )
				{
				osErr << "Error parsing expression: " << 
					Pi3String_getPtr( pError ) << ends;
				CONFIG_ERR( Object(), osErr.str() );
				Pi3Expression_delete( pExpr );
				pExpr = 0;
				Pi3String_delete( pError );
				return 0;
				};
			Pi3String_delete( pError );
			}
		else if ( !PIUtil_stricmp( KEY_CONF_FILE, pVariable ) )
			{
			Pi3String *pString = Pi3String_new( sFilePath );
			HTTPCore_relativeToAbsolutePath( PIObject_getDB( Object() ),
                pValue, pString );
			sFilePath = Pi3String_getPtr( pString );
			Pi3String_delete( pString );
			}
		else
			{
			osErr << "Unknown directive '" << pVariable <<
				"'" << ends;
			CONFIG_ERR( Object(), osErr.str() );
			return 0;
			};

		return 1;
		};

public:
	ExpressionLogger( PIObject *pObject, int iArgc, const char *ppArgv[] )
	:	HandlerBaseHTTP( pObject ),
		sFilePath(),
		sOpenMode( "w" ),
		pExpr( 0 )
		{
		ReadParameters( iArgc, ppArgv );
		if ( !IsOK() )
			{ return; };

		/* --- attempt to open the file --- */
		tFd = PIFile_open( sFilePath, sOpenMode );
		if ( tFd == PIPLATFORM_FD_INVALID )
			{
			PIOStrStream osErr;
			osErr << "ExpressionLogger: ";
			osErr << "Could not open file '" << sFilePath << 
				"', error '" << PIPlatform_getLastError() << "'" << ends;
			CONFIG_ERR( Object(), osErr.str() );
			SetOK( 0 );
			return;
			};
		};

	~ExpressionLogger()	
		{
		if ( tFd != PIPLATFORM_FD_INVALID )
			{ PIFile_close( tFd ); };
		Pi3Expression_delete( pExpr );
		};

	int Handle( int /* iPhase */, PIHTTP &tPIHTTP, PIIOBuffer &/* tBuffer */ )
		{
		assert( pExpr );
		if ( !pExpr )
			{ return PIAPI_ERROR; };

		enum { BUF_SIZE=1024 };
		char szBuf[BUF_SIZE];
		char *pBuf = szBuf;
		int iLen = Pi3Expression_write( pExpr, &tPIHTTP, 0, pBuf, BUF_SIZE );
		if ( iLen>BUF_SIZE )
			{
			pBuf = PI_NEW( char[iLen] );
			Pi3Expression_write( pExpr, &tPIHTTP, 0, pBuf, iLen );
			};
		
		/* --- write the message --- */
		int iRet = PIFile_writeAtomic( tFd, iLen, pBuf );
		if ( pBuf != szBuf )
			{
			PI_DELETE( [] pBuf );
			};
		if ( iRet )
			{
			/* --- an error occurred --- */
			HTTPCore_logError( &tPIHTTP, "ExpressionLogger: \
 PIFile_writeAtomic() failed for file '%s', error code '%d'.",
                (const char *)sFilePath, PIPlatform_getLastError() );
			return PIAPI_ERROR;
			};

		/* ---
		Return CONTINUE to allow other logging to happen
		--- */
		return PIAPI_CONTINUE;
		};
};

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int ExpressionLogger_constructor( PIObject *pObj,
	int iArgc, const char *ppArgv[] )
{
	return HandlerBaseHTTP_constructor( pObj, PI_NEW( ExpressionLogger( pObj,
		iArgc, ppArgv ) ) );
}

#if 0
/*___+++CNF_BEGIN+++___*/
	<Class>
		Name ExpressionLoggerClass
		Type LogicExtension
		Library HTTP
		OnClassLoad HandlerBaseHTTP_onClassLoad
		Constructor ExpressionLogger_constructor
		CopyConstructor HandlerBaseHTTP_copyConstructor
		Destructor HandlerBaseHTTP_destructor
		Execute HandlerBaseHTTP_execute
	</Class>

	<Object>
		Name ExpressionLogger
		Class ExpressionLoggerClass
	</Object>

/*___+++CNF_END+++___*/
#endif

