/*____________________________________________________________________________*\
 *

 Copyright (c) 1997-2003 John Roy, Holger Zimmermann. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. The name of the author may not be used to endorse or promote products
    derived from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/HTTP/RefFile.cpp,v $
 * $Date: 2003/05/13 18:42:03 $
 *
 Description:
	Handler for refusing a file based on its permissions compared with
	a supplied mask.

\*____________________________________________________________________________*/
//$SourceTop:$

#include <iostream.h>
#include <ctype.h>

#include "HandBase.h"
#include "HTTPCore.h"
#include "HTTPUtil.h"
#include "PIStrStr.h"
#include "DblList.h"

/*____________________________________________________________________________*\
 *
 Description:
\*____________________________________________________________________________*/
#define KEY_CONF_REFUSESTATUS	"RefuseStatus"
#define KEY_CONF_ALLOWFILEMASK	"AllowFileMask"
#define VALUE_EXISTS			'E'
#define VALUE_FILE				'F'
#define VALUE_LINK				'L'
#define VALUE_DIRECTORY			'D'
#define RF_EXISTS				0x01
#define RF_FILE					0x02
#define RF_LINK					0x04
#define RF_DIRECTORY			0x08

#if 0
	/*
	** HTML documentation for this handler
	*/
/*___+++HTMLDOC_BEGIN+++___*/
Name:
	RefuseFileByMask

Description:
	Set an HTTP error response code if the physical resource does not
	match particular criteria.	

Options:
<H5>Overview</H5>
<TABLE BORDER=1>
<TH>Option
<TH>Default
<TH>Values
<TH>Short Description
<TH>Example(s)

<TR>
<TD>RefuseStatus
<TD>-
<TD>HTTP Response code
<TD>Response code to set on failure
<TD>RefuseStatus=404; RefuseStatus="403"

<TR>
<TD>AllowFileMask
<TD>-
<TD>Combination of 'E', 'F', 'L', 'D'
<TD>Logical OR of condition to check
<TD>AllowFileMask="EF"; AllowFileMask="EFD"

</TABLE>
<STRONG>-</STRONG> in the <IT>default</IT> indicates no default<BR>
<STRONG>+</STRONG> in the <IT>default</IT> indicates the field is mandatory<BR>

<H4>Description of Options</H4>
<H5>
	RefuseStatus
</H5>
HTTP status code to set if the conditions specified by AllowFileMask are
not met. Typically this would be '404' for resource not found, or '403' for
forbidden.

<H5>
	AllowFileMask
</H5>
A group of letters indicating which properties of the physical resource
to check.

<CENTER>
<TABLE BORDER=1>
<TH>Option flag
<TH>Meaning

<TR>
<TD>E
<TD>Resource must exist

<TR>
<TD>F
<TD>Resource is a regular file

<TR>
<TD>D
<TD>Resource is a directory

<TR>
<TD>L
<TD>Resource is a symbolic link
</TABLE>
</CENTER>
These flags are not case sensitive.

Phase:
	CHECKPATH

Returns:
	INT_REDIRECT (via HTTPUtil_doHTTPError()) to cause the HTTP error
	specified by RefuseStatus to occur if the conditions specified by
	AllowFileMask are not met. Otherwise PIAPI_CONTINUE is returned to
	allow other path checking handlers to be invoked.
	PIAPI_ERROR is returned if this handler is used to respond to a
	phase other than 'CHECKPATH'.

Example:
	<PRE>
	&lt;Object&gt;
		Name RefuseFileByMask
		Class RefuseFileByMaskClass
	&lt;/Object&gt;

	&lt;Object&gt;
		...
		CheckPath RefuseFileByMask AllowFileMask="EF" RefuseStatus=404
		...
	&lt;/Object&gt;

	</PRE>
/*___+++HTMLDOC_END+++___*/
#endif

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class RefuseFileByMask : public HandlerBaseHTTP
{
private:
	/* ---
	Configuration data
	--- */
	int iStatus;
	int iAllowFlags;
	PIString sFileMask;

protected:
	int Parameter( const char *pVariable, const char *pValue,
		const char *pWhere )
		{
		assert( pVariable && pValue );
		PIOStrStream os;
		os << pWhere << "RefuseFileByMask: ";
		if ( !PIUtil_stricmp( KEY_CONF_REFUSESTATUS, pVariable ) )
			{
			iStatus = atoi( pValue );
			}
		else if ( !PIUtil_stricmp( KEY_CONF_ALLOWFILEMASK, pVariable ) )
			{
			int iLen = strlen( pValue );
			for( int i=0; i<iLen; i++ )
				{
				char c = toupper( pValue[i] );
				switch( c )
					{
					case VALUE_EXISTS:
						iAllowFlags |= RF_EXISTS; break; 
					case VALUE_FILE:
						iAllowFlags |= RF_FILE; break; 
					case VALUE_LINK:
						iAllowFlags |= RF_LINK; break; 
					case VALUE_DIRECTORY:
						iAllowFlags |= RF_DIRECTORY; break; 
					default:;
						os << "Unknown mask flag '" << c << 
							"'" << ends;
						CONFIG_ERR( Object(), os.str() );
						return 0;
					};
				};

			/* --- save the mask for error logging --- */
			sFileMask = pValue;
			}
		else
			{
			os << "Unknown directive '" << pVariable <<
				"'" << ends;
			CONFIG_ERR( Object(), os.str() );
			return 0;
			};

		return 1;
		};

public:
	RefuseFileByMask( PIObject *pObject, int iArgc, const char *ppArgv[] )
	:	HandlerBaseHTTP( pObject ), iStatus( 0 ), iAllowFlags( 0 )
		{
		ReadParameters( iArgc, ppArgv );
		if ( iStatus==0 )
			{
			CONFIG_ERR( Object(), "RefuseFileByMask: 'Status' not \
defined" );
			SetOK( 0 );
			return;
			};
		if ( iAllowFlags==0 )
			{
			CONFIG_ERR( Object(), "RefuseFileByMask: No mask given, \
handler has no effect" );
			SetOK( 0 );
			return;
			};
		};

	int Handle( int iPhase, PIHTTP &tPIHTTP, PIIOBuffer & )
		{
		assert( IsOK() );
		if ( iPhase!=PH_CHECKPATH ) { return PIAPI_ERROR; };

		const char *pPath = (const char *)PIDB_lookup( tPIHTTP.pResponseDB,
			PIDBTYPE_STRING, KEY_INT_PATH, 0 );
		PIFInfo *pFInfo = HTTPCore_getCachedFile( pPath ); 
		if ( !pFInfo ) { return PIAPI_ERROR; };

		/* --- reason for refusal --- */
		const char *pReason = 0;

		/* --- check for existance --- */
		if ( ( iAllowFlags & RF_EXISTS ) && !PIFInfo_exists( pFInfo ) )
			{
			pReason = "File does not exist";
			goto do_refuse;
			};

		if ( ( iAllowFlags & RF_FILE  ) && PIFInfo_isRegular( pFInfo ) )
			{
			goto do_allow;
			}
		else if ( ( iAllowFlags & RF_LINK ) && PIFInfo_isLink( pFInfo ) )
			{
			goto do_allow;
			}
		else if ( ( iAllowFlags & RF_DIRECTORY ) &&
				PIFInfo_isDirectory( pFInfo ) )
			{
			goto do_allow;
			};

		/* --- file does not match the mask --- */
		pReason = "File does not match required mask";

do_refuse:
		/* --- log an error message --- */
		HTTPCore_logError( &tPIHTTP, "RefuseFileByMask: File with \
path '%s' refused. Required filemask is '%s'. Reason: %s. \
Response status set to '%d'.",  
			pPath,
			(const char *)sFileMask,
			pReason,
			iStatus
			);

		/* --- release the file --- */
		HTTPCore_releaseCachedFile( pFInfo );

		/* --- cause error --- */
		return HTTPUtil_doHTTPError( &tPIHTTP, iStatus );

do_allow:
		HTTPCore_releaseCachedFile( pFInfo );
		return PIAPI_CONTINUE;
		};
};

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int RefuseFileByMask_constructor( PIObject *pObj,
	int iArgc, const char *ppArgv[] )
{
	return HandlerBaseHTTP_constructor( pObj, PI_NEW( RefuseFileByMask( pObj,
		iArgc, ppArgv ) ) );
}

#if 0
/*___+++CNF_BEGIN+++___*/
	<Class>
		Name RefuseFileByMaskClass
		Type LogicExtension
		Library HTTP
		OnClassLoad HandlerBaseHTTP_onClassLoad
		Constructor RefuseFileByMask_constructor
		CopyConstructor HandlerBaseHTTP_copyConstructor
		Destructor HandlerBaseHTTP_destructor
		Execute HandlerBaseHTTP_execute
	</Class>

	<Object>
		Name RefuseFileByMask
		Class RefuseFileByMaskClass
	</Object>

/*___+++CNF_END+++___*/
#endif

