/*____________________________________________________________________________*\
 *

 Copyright (c) 1997-2003 John Roy, Holger Zimmermann. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. The name of the author may not be used to endorse or promote products
    derived from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/HTTP/SendSSI.cpp,v $
 * $Date: 2003/06/06 05:58:43 $
 *
 Description:
	Server-side include handling.

\*____________________________________________________________________________*/
//$SourceTop:$

#include <iostream.h>
#include <stdio.h>
#include <ctype.h>

#include "HandBase.h"
#include "HTTPCore.h"
#include "HTTPUtil.h"
#include "PIStrStr.h"
#include "Pi3Expr.h"

/*____________________________________________________________________________*\
 *
 Description:
\*____________________________________________________________________________*/
#define KEY_CONF_EXEC				"Exec"
#define KEY_VALUE_YES				"Yes"
#define KEY_CONF_HEADERPATTERN		"HeaderPattern"
#define KEY_CONF_FOOTERPATTERN		"FooterPattern"
#define KEY_CONF_EXECFILEFIXUP		"ExecFileFixup"

/*____________________________________________________________________________*\
 *
 Description:
\*____________________________________________________________________________*/
#if 0
	/*
	** HTML documentation for this handler
	*/
/*___+++HTMLDOC_BEGIN+++___*/
Name:
	SendSSI

Description:
	Parse and send server-side includes.

Options:
<H5>Overview</H5>
<TABLE BORDER=1>
<TH>Option
<TH>Default
<TH>Values
<TH>Short Description
<TH>Example(s)

<TR>
<TD>Exec
<TD>No
<TD>'Yes' or 'No'
<TD>Specific whether CGI scripts may be executed
<TD>Exec="Yes"

<TR>
<TD>HeaderPattern
<TD>-
<TD>A Pi3Expression
<TD>Output first before the page
<TD>HeaderPattern="<HTML><TITLE>The Title</TITLE>"

<TR>
<TD>FooterPattern
<TD>-
<TD>A Pi3Expression
<TD>Output last
<TD>FooterPattern="</HTML>"

<TR>
<TD>ExecFileFixup
<TD>-
<TD>A Pi3Expression
<TD>Evaluated to fixup CGI markers
<TD>ExecFileFixup="&dbreplace(response,string,ObjectMap,Scripts)"


</TABLE>
<STRONG>-</STRONG> in the <IT>default</IT> indicates no default<BR>
<STRONG>+</STRONG> in the <IT>default</IT> indicates the field is mandatory<BR>

<H4>Description of Options</H4>
<H5>
	Exec
</H5>
Specify whether or not CGI scripts may be executed as part of SSI processing.
Comparision with 'Yes' or 'No' is not case sensitive.

<H5>
	HeaderPattern
</H5>
Specifies a pattern of data that will be pre-appended to the file.

<H5>
	FooterPattern
</H5>
Specifies a pattern of data that will be appended to the file.

<H5>
	ExecFileFixup
</H5>
For <I>exec</I> ssi directives with the file specified via
'file=' this pattern is evaluated to allow a user defined
Pi3Expression to cause markers to be set. This is because
virtual path processing would normally set markers to indicate
which CGI handler to use, but 'file=' bypasses the virtual
path mapping phase.

Note:
Phase:
	HANDLE

Returns:
	PIAPI_COMPLETED if the file was sent.
	If this handler was invoked for a phase other than HANDLE then PIAPI_ERROR
	is returned.

Example:
	<PRE>
	&lt;Object&gt;
		Name SendSSI
		Class SendSSIClass
	&lt;/Object&gt;

	&lt;Object&gt;
		...
		Handle SendSSI Exec="Yes" 
		...
	&lt;/Object&gt;

	</PRE>
/*___+++HTMLDOC_END+++___*/
#endif

/*____________________________________________________________________________*\
 *
 Class:
 Description:
	class for storing SSI configuration values 
\*____________________________________________________________________________*/
class SSIConfig
{
public:
	PIString sTimeFmt;
	PIString sSizeFmt;
	PIString sErrMsg;

	SSIConfig()
	:	
		sTimeFmt( "%a, %d %b %Y %T" ),
		sSizeFmt( "%d bytes" ),
		sErrMsg( "** SSI Error **" )
		{};

	SSIConfig( const SSIConfig &tConfig )
	:
		sTimeFmt( tConfig.sTimeFmt ),
		sSizeFmt( tConfig.sSizeFmt ),
		sErrMsg( tConfig.sErrMsg )
		{};
};

/*____________________________________________________________________________*\
 *
 Class:
 Description:
	simple class to pass parameters to expression callback functions
\*____________________________________________________________________________*/
class SSIParam
{
public:
	PIFInfo *pFInfo;
	SSIConfig &tConfig;
	Pi3Expression *pExecFileFixup;
	SSIParam( PIFInfo *pTheFInfo, SSIConfig &tTheConfig, Pi3Expression *pExpr )
	:	pFInfo( pTheFInfo ), tConfig( tTheConfig ), pExecFileFixup( pExpr )
		{};
};

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
static int Internal_GMTTime( PIHTTP * /* pPIHTTP */, SSIConfig &tConfig,
	time_t tT, char *pszBuffer, int iLength )
{
	enum { BUF_SIZE=255 };
	char szBuf[BUF_SIZE+1];
	PIPlatform_beforeUnsafeBlock();
	struct tm *pTms = gmtime( &tT );
	int iRet = strftime( szBuf, BUF_SIZE, tConfig.sTimeFmt, pTms );
	PIPlatform_afterUnsafeBlock();
	szBuf[iRet] = '\0';
	if ( iLength )
		{ strncpy( pszBuffer, szBuf, iLength ); };
	return iRet;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
	Date GMT
\*____________________________________________________________________________*/
static int f_a( PIHTTP *pPIHTTP, void *pData, char *pszBuffer, int iLength )
{
	(void)*pPIHTTP;
	assert( pData );
	SSIParam *pParam = (SSIParam *)pData;
	time_t tT;
	time( &tT );
	return Internal_GMTTime( pPIHTTP, pParam->tConfig, tT, pszBuffer, 
		iLength );
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
	Date local
\*____________________________________________________________________________*/
static int f_b( PIHTTP *pPIHTTP, void *pData, char *pszBuffer, int iLength )
{
	(void)pPIHTTP;
	assert( pData );
	SSIParam *pParam = (SSIParam *)pData;
	time_t tT;
	time( &tT );
	enum { BUF_SIZE=255 };
	char szBuf[BUF_SIZE+1];
	PIPlatform_beforeUnsafeBlock();
	struct tm *pTms = localtime( &tT );
	int iRet = strftime( szBuf, BUF_SIZE, pParam->tConfig.sTimeFmt, pTms );
	PIPlatform_afterUnsafeBlock();
	szBuf[iRet] = '\0';
	if ( iLength )
		{ strncpy( pszBuffer, szBuf, iLength ); };
	return iRet;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
	Last modified
\*____________________________________________________________________________*/
static int f_c( PIHTTP *pPIHTTP, void *pData, char *pszBuffer, int iLength )
{
	(void)pPIHTTP;
	assert( pData );
	SSIParam *pParam = (SSIParam *)pData;
	time_t tT = PIFInfo_getLastModified( pParam->pFInfo );
	return Internal_GMTTime( pPIHTTP, pParam->tConfig, tT, pszBuffer, 
		iLength );
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
	Last modified
\*____________________________________________________________________________*/
static int f_d( PIHTTP *pPIHTTP, void *pData, char *pszBuffer, int iLength )
{
	(void)pPIHTTP;
	(void)pData;
#define TMP "## Not Implemented ##"
	if ( iLength>0 )
		{
		assert( pszBuffer );
		strncpy( pszBuffer, TMP, iLength );
		};
	return sizeof(TMP)-1;
}

/*____________________________________________________________________________*\
 *
 Description:
	Create a map of FnPi3Write functions for parameters to expressions
\*____________________________________________________________________________*/
static FnPi3Write aFunctions[256] =
{
/*
0   1   2   3   4   5   6   7   8   9  10  11  12  13  14  15
*/
0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 	/* 0 .. 15 */
0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 	/* 16 .. 31 */
0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 	/* 32 .. 47 */
0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 	/* 48 .. 63 */
/*  A   B   C   D   E   F   G   H   I   J   K   L   M   N   O  */
0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  /* 64 .. 79 */	
0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 	/* 80 .. 95 */
/*  a   b   c   d   e   f   g   h   i   j   k   l   m   n   o  */
0,f_a,f_b,f_c,f_d,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 	/* 96 .. 111 */
/*  q   r   s   t   u   v   w   x   y   z   */
0,	0,	0,	0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  /* 112 .. 127 */	
0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 	
0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 	
0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 	
0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 	
0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 	
0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 	
0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 	
0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0 	/* .. 255 */	
};

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class SendSSI : public HandlerBaseHTTP
{
private:
	/* --- function and DB of functions --- */
	typedef int (* SSIFunc)( SendSSI &, PIHTTP &, const char *, int,
			const char *, int, SSIParam & );
	PIDB *pFunctions;
	PIDB *pVariables;

	/* ---
	Configuration data
	--- */
	SSIConfig tConfig;
	int iAllowExec;
	Pi3Expression *pHeaderPattern;	/* pattern for top */
	Pi3Expression *pFooterPattern;	/* pattern for footer */
	Pi3Expression *pExecFileFixup;	/* pattern exec file fix up */

	/* ---
	Internal methods 
	--- */

	/* ---
	Dispatch a subrequest to get information on a file or return it to	
	the client.
	Returns non-zero on success, 0 on failure.
	--- */
	int GetSSIFile( PIHTTP &tPIHTTP, const char *pSSIFile,
		int iSSIFileLen, int iVirtual, SSIParam &tParam, int iExec,
		int iDoHandle, PIFInfo **ppFInfo )
		{
		/* --- make sub request context --- */
		PIHTTP *pChildHTTP = PIHTTP_newChild( &tPIHTTP );

        /*
        ** Find the '?' in the path (for Query String)
        */
        int i=0;
        for( ; i<iSSIFileLen && pSSIFile[i]!='?'; i++ );

		/* --- set path --- */
		PIString sTmp( pSSIFile, i );

        /* --- set query string --- */
        PIString sQS;
        if ( i<iSSIFileLen )
            { 
            i++;
            PIString sTmpQS( &(pSSIFile[i]), iSSIFileLen-i );
            sQS = sTmpQS;
            };

		if ( !iVirtual )
			{
			PIString sFile( PIFInfo_getPathRoot( tParam.pFInfo ) );
			sFile.Concatenate( HTTP_DIRSEPERATOR );
			sFile.Concatenate( sTmp );
			sTmp = sFile;
			};

		PIDB_replace( pChildHTTP->pResponseDB, PIDBTYPE_STRING,
			KEY_INT_PATH, (void *)(const char *)sTmp, 0 );
		PIDB_replace( pChildHTTP->pRequestDB, PIDBTYPE_STRING,
			KEY_HTTP_QUERYSTRING, (void *)(const char *)sQS, 0 );

		/* --- dispatch the sub request across a number of phases --- */
		int iRet = PIAPI_ERROR;
		iRet = HTTPCore_dispatch(
			pChildHTTP,
			iVirtual ? PH_MAPPING : PH_CHECKPATH,
			PH_CHECKTYPE );
			
		/* --- check for error --- */
		if ( iRet!=PIAPI_COMPLETED ||
			((pChildHTTP->iStatus!=0) && (pChildHTTP->iStatus!=ST_OK)) )
			{
			PIHTTP_delete( pChildHTTP );
			return 0;
			};

		/* ---
		If this is exec and not virtual and we have a pExecFileFixup pattern
		execute it now to setup variables so the physical path is
		processed by the correct CGI handler. 

		This is necessary because virtual paths usually sets the
		appropriate markers as part of path mapping
		--- */
		if ( iExec && !iVirtual && tParam.pExecFileFixup )
			{
			Pi3Expression_write( tParam.pExecFileFixup, pChildHTTP, 0, 0, 0 );
			};

		/* ---
		check this is not an executable file, just check Content-Type is
		not 'internal/x-cgi'.
		--- */
		const char *pContentType = (const char *)PIDB_lookup(
			pChildHTTP->pResponseDB, PIDBTYPE_RFC822,
			KEY_HTTP_CONTENTTYPE, 0 );
		int iIsExec = (!pContentType ||
			!PIUtil_stricmp( pContentType, TYPE_MAGIC_CGI ) );
		if ( ( iIsExec && !iExec ) ||
			( !iIsExec && iExec ) )
			{
			PIHTTP_delete( pChildHTTP );
			return 0;
			};

		/* --- handle this request? --- */
		if ( iDoHandle )
			{
			/* --- handle this request --- */
			iRet = HTTPCore_dispatch( pChildHTTP, PH_HANDLE, PH_HANDLE );
			PIHTTP_delete( pChildHTTP );
			return iRet==PIAPI_COMPLETED ? 1 : 0;
			}
		else
			{
			/* --- get mapped path --- */
			const char *pPath = (const char *)PIDB_lookup(
				pChildHTTP->pResponseDB, PIDBTYPE_STRING,
				KEY_INT_PATH, 0 );

			assert( ppFInfo );
			if ( !ppFInfo )
				{ return 0; };
			*ppFInfo = HTTPCore_getCachedFile( pPath ); 
			PIHTTP_delete( pChildHTTP );
			return 1;
			};
		};

	/* --- 'config' directive --- */
	static int f_config( SendSSI &/* tThis */, PIHTTP &tPIHTTP,
			const char *pVar, int iVarLen, const char *pVal, int iValLen,
			SSIParam &tParam )
		{
		if ( !pVar )
			{}
		else if ( iVarLen==7 && !PIUtil_strncmpi( pVar, "timefmt", 7 ) )
			{
			PIString sTmp( pVal, iValLen );
			tParam.tConfig.sTimeFmt = sTmp;
			return 1;
			}
		else if ( iVarLen==7 && !PIUtil_strncmpi( pVar, "sizefmt", 7 ) )
			{
			PIString sTmp( pVal, iValLen );
			tParam.tConfig.sSizeFmt = sTmp;
			return 1;
			}
		else if ( iVarLen==6 && !PIUtil_strncmpi( pVar, "errmsg", 6 ) )
			{
			PIString sTmp( pVal, iValLen );
			tParam.tConfig.sErrMsg = sTmp;
			return 1;
			};

		/* --- error, no variable, or unknown variable --- */
		PIIOBuffer_write(
			tPIHTTP.pBuffer, 
			tParam.tConfig.sErrMsg,
			tParam.tConfig.sErrMsg.Len(),
			PIIOBUF_NONE );

		return 0;
		};

	/* --- 'echo' directive --- */
	static int f_echo( SendSSI &tThis, PIHTTP &tPIHTTP, const char *pVar,
			int iVarLen, const char *pVal, int iValLen,
			SSIParam &tParam )
		{
		if ( !pVar || iVarLen!=3 || PIUtil_strncmpi( pVar, "var", 3 ) )
			{
			PIIOBuffer_write(
				tPIHTTP.pBuffer, 
				tParam.tConfig.sErrMsg,
				tParam.tConfig.sErrMsg.Len(),
				PIIOBUF_NONE );
			return 0;
			};

		Pi3Expression *pExpr = 0;
		if ( pVal )
			{
			PIString sTmp( pVal, iValLen );
			pExpr = (Pi3Expression *)PIDB_lookup( tThis.pVariables,
				PIDBTYPE_OPAQUE, sTmp, 0 );
			};

		if ( !pExpr )
			{
			/* --- invalid variable --- */
			PIIOBuffer_write(
				tPIHTTP.pBuffer, 
				tParam.tConfig.sErrMsg,
				tParam.tConfig.sErrMsg.Len(),
				PIIOBUF_NONE );
			
			return 0;
			};

		/* --- write out variable --- */
		enum { BUF_SIZE=1024 };
		char szBuf[ BUF_SIZE ];
		int iLen = Pi3Expression_write( pExpr, &tPIHTTP, &tParam, szBuf, 
			BUF_SIZE );
		if ( iLen>BUF_SIZE )
			{
			/* --- buffer was not big enough, try again --- */
			char *pBuf = (char *)PIHTTP_allocMem( &tPIHTTP, iLen );
			Pi3Expression_write( pExpr, &tPIHTTP, &tParam, pBuf, iLen );
			PIIOBuffer_write(
				tPIHTTP.pBuffer, 
				pBuf,
				iLen,
				PIIOBUF_NONE );
			}
		else
			{
			PIIOBuffer_write(
				tPIHTTP.pBuffer, 
				szBuf,
				iLen,
				PIIOBUF_NONE );
			};
	
		return 1;
		};

	/* --- 'flastmod' last modification time of a file --- */
	static int f_flastmod( SendSSI &tThis, PIHTTP &tPIHTTP,
			const char *pVar, int iVarLen, const char *pVal, int iValLen,
			SSIParam &tParam )
		{
		enum { ERROR=-1, FILE=0, VIRTUAL=1 } iType = ERROR;
		if ( !pVar )
			{}
		else if ( iVarLen==7 && !PIUtil_strncmpi( pVar, "virtual", 7 ) )
			{ iType = VIRTUAL; }
		else if ( iVarLen==4 && !PIUtil_strncmpi( pVar, "file", 4 ) )
			{ iType = FILE; };

		if ( iType!=ERROR )
			{
			PIFInfo *pFile;
			if ( tThis.GetSSIFile( tPIHTTP, pVal, iValLen, iType,
				tParam, 0, 0, &pFile  ) )
				{
				enum { BUF_SIZE=1024 };
				char szBuf[ BUF_SIZE ];
				time_t tT = PIFInfo_getLastModified( pFile );
				HTTPCore_releaseCachedFile( pFile );
				int iLen = Internal_GMTTime( &tPIHTTP, tParam.tConfig, tT,
					szBuf, BUF_SIZE );
				if ( iLen>BUF_SIZE )
					{
					assert( 0 );		/* should never really happen */
					iLen = BUF_SIZE;
					};
				PIIOBuffer_write(
					tPIHTTP.pBuffer, 
					szBuf,
					iLen,
					PIIOBUF_NONE );
				return 1;
				};
			};

		/* --- error, no variable, or unknown variable --- */
		PIIOBuffer_write(
			tPIHTTP.pBuffer, 
			tParam.tConfig.sErrMsg,
			tParam.tConfig.sErrMsg.Len(),
			PIIOBUF_NONE );

		return 0;
		};

	/* --- 'fsize' size of a file --- */
	static int f_fsize( SendSSI &tThis, PIHTTP &tPIHTTP, const char *pVar,
			int iVarLen, const char *pVal, int iValLen,
			SSIParam &tParam )
		{
		enum { ERROR=-1, FILE=0, VIRTUAL=1 } iType = ERROR;
		if ( !pVar )
			{}
		else if ( iVarLen==7 && !PIUtil_strncmpi( pVar, "virtual", 7 ) )
			{ iType = VIRTUAL; }
		else if ( iVarLen==4 && !PIUtil_strncmpi( pVar, "file", 4 ) )
			{ iType = FILE; };

		if ( iType!=ERROR )
			{
			PIFInfo *pFile;
			if ( tThis.GetSSIFile( tPIHTTP, pVal, iValLen, iType,
				tParam, 0, 0, &pFile ) )
				{
				enum { BUF_SIZE=1024 };
				char szBuf[ BUF_SIZE ];
				sprintf( szBuf, "%lu", PIFInfo_getSize( pFile ) );
				HTTPCore_releaseCachedFile( pFile );
				int iLen = strlen( szBuf );
				PIIOBuffer_write(
					tPIHTTP.pBuffer, 
					szBuf,
					iLen,
					PIIOBUF_NONE );
				return 1;
				};
			};

		/* --- error, no variable, or unknown variable --- */
		PIIOBuffer_write(
			tPIHTTP.pBuffer, 
			tParam.tConfig.sErrMsg,
			tParam.tConfig.sErrMsg.Len(),
			PIIOBUF_NONE );

		return 0;
		};

	/* --- 'include' include a file --- */
	static int f_include( SendSSI &tThis, PIHTTP &tPIHTTP,
			const char *pVar, int iVarLen, const char *pVal, int iValLen,
			SSIParam &tParam )
		{
		enum { ERROR=-1, FILE=0, VIRTUAL=1 } iType = ERROR;
		if ( !pVar )
			{}
		else if ( iVarLen==7 && !PIUtil_strncmpi( pVar, "virtual", 7 ) )
			{ iType = VIRTUAL; }
		else if ( iVarLen==4 && !PIUtil_strncmpi( pVar, "file", 4 ) )
			{ iType = FILE; };

		/* ---
		Include file
		--- */
		if ( iType!=ERROR )
			{
			if ( tThis.GetSSIFile( tPIHTTP, pVal, iValLen, iType,
				tParam, 0, 1, 0 ) )
				{ return 1; };
			};

		/* --- error, no variable, or unknown variable --- */
		PIIOBuffer_write(
			tPIHTTP.pBuffer, 
			tParam.tConfig.sErrMsg,
			tParam.tConfig.sErrMsg.Len(),
			PIIOBUF_NONE );

		return 0;
		};

	/* --- 'exec' execute a cgi program --- */
	static int f_exec( SendSSI &tThis, PIHTTP &tPIHTTP,
			const char *pVar, int iVarLen, const char *pVal, int iValLen,
			SSIParam &tParam )
		{
		enum { ERROR=-1, FILE=0, VIRTUAL=1 } iType = ERROR;
		if ( !pVar )
			{}
		else if ( iVarLen==7 && !PIUtil_strncmpi( pVar, "virtual", 7 ) )
			{ iType = VIRTUAL; }
		else if ( iVarLen==4 && !PIUtil_strncmpi( pVar, "file", 4 ) )
			{ iType = FILE; };

		/* ---
		Exec file
		--- */
		if ( iType!=ERROR && tThis.iAllowExec )
			{
			/* ---
			Do the stuff
			--- */
			if ( tThis.GetSSIFile( tPIHTTP, pVal, iValLen, iType,
				tParam, 1, 1, 0 ) )
				{ return 1; };
			};

		/* --- error, no variable, or unknown variable --- */
		PIIOBuffer_write(
			tPIHTTP.pBuffer, 
			tParam.tConfig.sErrMsg,
			tParam.tConfig.sErrMsg.Len(),
			PIIOBUF_NONE );

		return 0;
		};

	/* --- 'pi3extn' directive (the obligatory proprietory extension) --- */
	static int f_pi3ext( SendSSI &/* tThis */, PIHTTP &tPIHTTP,
			const char *pVar, int iVarLen, const char *pVal, int iValLen,
			SSIParam &tParam )
		{
		if ( !pVar )
			{}
		else if ( iVarLen==4 && !PIUtil_strncmpi( pVar, "expr", 4 ) )
			{
			PIString sTmp( pVal, iValLen );
			Pi3String *pError = Pi3String_new( 0 );
			Pi3Expression *pTmpExpr = Pi3Expression_new( sTmp, aFunctions,
				pError );
			Pi3String_delete( pError );
			if ( pTmpExpr )
				{
				enum { BUF_SIZE=1024 };
				char szBuf[ BUF_SIZE ];
				int iLen = Pi3Expression_write( pTmpExpr, &tPIHTTP, 0, szBuf,
					BUF_SIZE );
				if ( iLen>BUF_SIZE )
					{
					/* --- buffer was not big enough, try again --- */
					char *pBuf = (char *)PIHTTP_allocMem( &tPIHTTP, iLen );
					Pi3Expression_write( pTmpExpr, &tPIHTTP, 0, pBuf, iLen );
					PIIOBuffer_write( tPIHTTP.pBuffer, pBuf, iLen,
						PIIOBUF_NONE );
					}
				else
					{
					PIIOBuffer_write( tPIHTTP.pBuffer, szBuf, iLen,
						PIIOBUF_NONE );
					};
				Pi3Expression_delete( pTmpExpr );
				return 1;
				};
			};

		/* --- some error --- */
		PIIOBuffer_write(
			tPIHTTP.pBuffer, 
			tParam.tConfig.sErrMsg,
			tParam.tConfig.sErrMsg.Len(),
			PIIOBUF_NONE );

		return 0;
		};

protected:
	/*___ +++++++++++++++++++++++++++++++++++++++++++++++++ ___ *

				Load an expression or condition

	 *___ +++++++++++++++++++++++++++++++++++++++++++++++++ ___ */
	int LoadExpression( const char *pName, Pi3Expression **ppExpression,
		FnPi3Write *pFunctions, const char *pValue, PIOStrStream &os )	
		{
		assert( ppExpression );
		Pi3Expression *pExpression = *ppExpression;
		if ( pExpression )
			{
			os << "'" << pName << "' may only be specified once." << ends;
			CONFIG_ERR( Object(), os.str() );
			return 0;
			};
		Pi3String *pError = Pi3String_new( 0 );
		pExpression = Pi3Expression_new( pValue, pFunctions, pError );
		if ( !pExpression )
			{
			os << "Error parsing expression: " <<
				Pi3String_getPtr( pError ) << ends;
			CONFIG_ERR( Object(), os.str() );
			Pi3Expression_delete( pExpression );
			Pi3String_delete( pError );
			return 0;
			};
		Pi3String_delete( pError );
		*ppExpression = pExpression;
		return 1;
		};

	int Parameter( const char *pVariable, const char *pValue,
		const char *pWhere )
		{
		assert( pVariable && pValue );
		PIOStrStream os;
		os << pWhere << "SendSSI: ";
		if ( !PIUtil_stricmp( KEY_CONF_EXEC, pVariable ) )
			{
			if ( !PIUtil_stricmp( pValue, KEY_VALUE_YES ) )
				{ iAllowExec = 1; };
			}
		else if ( !PIUtil_stricmp( KEY_CONF_HEADERPATTERN, pVariable ) )
			{
			if ( !LoadExpression( KEY_CONF_HEADERPATTERN, &pHeaderPattern,
				0, pValue, os ) )
				{ return 0; };
			assert( pHeaderPattern );
			}
		else if ( !PIUtil_stricmp( KEY_CONF_FOOTERPATTERN, pVariable ) )
			{
			if ( !LoadExpression( KEY_CONF_FOOTERPATTERN, &pFooterPattern,
				0, pValue, os ) )
				{ return 0; };
			assert( pFooterPattern );
			}
		else if ( !PIUtil_stricmp( KEY_CONF_EXECFILEFIXUP, pVariable ) )
			{
			if ( !LoadExpression( KEY_CONF_EXECFILEFIXUP, &pExecFileFixup,
				0, pValue, os ) )
				{ return 0; };
			assert( pExecFileFixup );
			}
		else
			{
			os << "Unknown directive '" << pVariable <<
				"'" << ends;
			CONFIG_ERR( Object(), os.str() );
			return 0;
			};

		return 1;
		};

public:
	SendSSI( PIObject *pObject, int iArgc, const char *ppArgv[] )
	:
		HandlerBaseHTTP( pObject ),
		pFunctions( PIDB_new( 0, "SSI Functions" ) ),
		pVariables( PIDB_new( 0, "SSI Variables" ) ),
		iAllowExec( 0 ),
		pHeaderPattern( 0 ),
		pFooterPattern( 0 ),
		pExecFileFixup( 0 )
		{
		ReadParameters( iArgc, ppArgv );

		if ( !IsOK() )
			{ return; };

		/* --- insert SSI directives --- */
		PIDB_add( pFunctions, PIDBTYPE_OPAQUE, "CONFIG", (void *)f_config, 0 );
		PIDB_add( pFunctions, PIDBTYPE_OPAQUE, "ECHO", (void *)f_echo, 0 );
		PIDB_add( pFunctions, PIDBTYPE_OPAQUE, "FLASTMOD", (void *)f_flastmod, 0 );
		PIDB_add( pFunctions, PIDBTYPE_OPAQUE, "FSIZE", (void *)f_fsize, 0 );
		PIDB_add( pFunctions, PIDBTYPE_OPAQUE, "INCLUDE", (void *)f_include, 0 );
		PIDB_add( pFunctions, PIDBTYPE_OPAQUE, "EXEC", (void *)f_exec, 0 );
		PIDB_add( pFunctions, PIDBTYPE_OPAQUE, "PI3EXT", (void *)f_pi3ext, 0 );

		/* ---
		insert SSI variables

		Each variable is mapped to a Pi3Expression. Not the fastest, but
		pretty convienient.

		--- */
		Pi3String *pDummy = Pi3String_new( 0 );
	
		PIDB_add( pVariables, PIDBTYPE_OPAQUE, "AUTH_TYPE", 
			Pi3Expression_new( "$x", aFunctions, pDummy ), 0 );
		PIDB_add( pVariables, PIDBTYPE_OPAQUE, "CONTENT_LENGTH", 
			Pi3Expression_new( "$Y", aFunctions, pDummy ), 0 );
		PIDB_add( pVariables, PIDBTYPE_OPAQUE, "CONTENT_TYPE", 
			Pi3Expression_new( "$C", aFunctions, pDummy ), 0 );
		PIDB_add( pVariables, PIDBTYPE_OPAQUE, "DOCUMENT_NAME", 
			Pi3Expression_new( "$f", 0, pDummy ), 0 );
		PIDB_add( pVariables, PIDBTYPE_OPAQUE, "DOCUMENT_URI", 
			Pi3Expression_new( "$U", 0, pDummy ), 0 );
		PIDB_add( pVariables, PIDBTYPE_OPAQUE, "HTTPS", 
			Pi3Expression_new( "$g", 0, pDummy ), 0 );
		PIDB_add( pVariables, PIDBTYPE_OPAQUE, "HTTPS_KEYSIZE", 
			Pi3Expression_new( "$K", 0, pDummy ), 0 );
		PIDB_add( pVariables, PIDBTYPE_OPAQUE, "HTTPS_SECRETKEYSIZE", 
			Pi3Expression_new( "$G", 0, pDummy ), 0 );
		PIDB_add( pVariables, PIDBTYPE_OPAQUE, "QUERY_STRING", 
			Pi3Expression_new( "$q", 0, pDummy ), 0 );
		PIDB_add( pVariables, PIDBTYPE_OPAQUE, "SERVER_ADMIN", 
			Pi3Expression_new( "$a", 0, pDummy ), 0 );
		PIDB_add( pVariables, PIDBTYPE_OPAQUE, "SERVER_SOFTWARE", 
			Pi3Expression_new( "$S", 0, pDummy ), 0 );
		PIDB_add( pVariables, PIDBTYPE_OPAQUE, "SERVER_NAME", 
			Pi3Expression_new( "$v", 0, pDummy ), 0 );
		PIDB_add( pVariables, PIDBTYPE_OPAQUE, "SERVER_PORT", 
			Pi3Expression_new( "$p", 0, pDummy ), 0 );
		PIDB_add( pVariables, PIDBTYPE_OPAQUE, "SERVER_PROTOCOL", 
			Pi3Expression_new( "&dblookup(request,string,Protocol)",
			0, pDummy ), 0 );
		PIDB_add( pVariables, PIDBTYPE_OPAQUE, "REQUEST_METHOD", 
			Pi3Expression_new( "$m", 0, pDummy ), 0 );
		PIDB_add( pVariables, PIDBTYPE_OPAQUE, "REMOTE_ADDR", 
			Pi3Expression_new( "$A", 0, pDummy ), 0 );
		PIDB_add( pVariables, PIDBTYPE_OPAQUE, "REMOTE_HOST", 
			Pi3Expression_new( "$h", 0, pDummy ), 0 );
		PIDB_add( pVariables, PIDBTYPE_OPAQUE, "REMOTE_IDENT", 
			Pi3Expression_new( "$i", 0, pDummy ), 0 );
		PIDB_add( pVariables, PIDBTYPE_OPAQUE, "REMOTE_USER", 
			Pi3Expression_new( "$u", 0, pDummy ), 0 );
		PIDB_add( pVariables, PIDBTYPE_OPAQUE, "HTTP_ACCEPT", 
			Pi3Expression_new( "&dblookup(request,rfc822,Accept)",
			0, pDummy ), 0 );
		PIDB_add( pVariables, PIDBTYPE_OPAQUE, "HTTP_USER_AGENT", 
			Pi3Expression_new( "&dblookup(request,rfc822,User-Agent)",
			0, pDummy ), 0 );
		PIDB_add( pVariables, PIDBTYPE_OPAQUE, "REFERER", 
			Pi3Expression_new( "&dblookup(request,rfc822,Referer)",
			0, pDummy ), 0 );
		PIDB_add( pVariables, PIDBTYPE_OPAQUE, "DATE_GMT", 
			Pi3Expression_new( "%a", aFunctions, pDummy ), 0 );
		PIDB_add( pVariables, PIDBTYPE_OPAQUE, "DATE_LOCAL", 
			Pi3Expression_new( "%b", aFunctions, pDummy ), 0 );
		PIDB_add( pVariables, PIDBTYPE_OPAQUE, "LAST_MODIFIED", 
			Pi3Expression_new( "%c", aFunctions, pDummy ), 0 );
		PIDB_add( pVariables, PIDBTYPE_OPAQUE, "PATH_INFO", 
			Pi3Expression_new( "$I", aFunctions, pDummy ), 0 );
		PIDB_add( pVariables, PIDBTYPE_OPAQUE, "PATH_TRANSLATED", 
			Pi3Expression_new( "$Z", aFunctions, pDummy ), 0 );
		PIDB_add( pVariables, PIDBTYPE_OPAQUE, "ERROR_MESSAGES", 
			Pi3Expression_new( "&dblookup(request,string,ErrorMessage,,,&_())",
			0, pDummy ), 0 );
		Pi3String_delete( pDummy );
		};

	~SendSSI()
		{
		/* --- delete functions DB --- */
		PIDB_delete( pFunctions );

		/* --- delete variables from variables DB --- */
		PIDBIterator *pIter = PIDB_getIterator( pVariables, PIDBTYPE_OPAQUE,
			0, 0 );
		if ( pIter )
			{
			for( ; PIDBIterator_atValidElement( pIter );
				PIDBIterator_next( pIter ) )
				{
				Pi3Expression_delete(
					(Pi3Expression *)PIDBIterator_current( pIter, 0 ) );
				};
			PIDBIterator_delete( pIter );
			};

		/* --- delete variables DB --- */
		PIDB_delete( pVariables );

		Pi3Expression_delete( pHeaderPattern );
		Pi3Expression_delete( pFooterPattern );
		Pi3Expression_delete( pExecFileFixup );
		};

	/* ---
	Process a single SSI directive
	Return 0 on error, 1 on success.
	--- */
	int ProcessSSIDirective( PIHTTP &tPIHTTP, const char *pDirective, 
		int iDirectiveLen, SSIParam &tParam )
		{
		/* --- get first word (command) in directive --- */
		int i;
		for( i=0; i<iDirectiveLen && !(isspace(pDirective[i])); i++);

		/* --- portions of directive --- */
		const char *pCommand = pDirective;
		int iCommandLen = i;
		const char *pVariable = 0;
		int iVariableLen = 0;
		const char *pValue = 0;
		int iValueLen = 0;
	
		if ( i<iDirectiveLen )
			{
			/* --- get variable --- */
			for( ; i<iDirectiveLen && (isspace(pDirective[i])); i++);
			pDirective = &( pDirective[i] ); iDirectiveLen -= i; i = 0;
			pVariable = pDirective;
			for( ; i<iDirectiveLen && !(isspace(pDirective[i])) &&
				pDirective[i]!='='; i++);
			iVariableLen = i;

			/* --- get value --- */
			for( ; i<iDirectiveLen && ( (isspace(pDirective[i])) ||
				( pDirective[i]=='=' ) ); i++);
			int iQuoted = 0;
			if ( pDirective[i]=='"' ) { iQuoted = 1; i++; };
			pDirective = &( pDirective[i] ); iDirectiveLen -= i; i = 0;
			for( ; i<iDirectiveLen && ( !(isspace(pDirective[i])) || iQuoted )
				; i++)
				{
				if ( pDirective[i]=='"' && iQuoted )
					{ break; };
				};
			pValue = pDirective;
			iValueLen = i;
			};

		/* --- get directive function and execute it --- */
		PIString sCommand( pCommand, iCommandLen );
		sCommand.ConvertToUpperCase();
		SSIFunc tFn = (SSIFunc)PIDB_lookup( pFunctions, PIDBTYPE_OPAQUE,
			sCommand, 0 );

		if ( !tFn )
			{
			/* --- unrecognised directive --- */
			PIIOBuffer_write(
				tPIHTTP.pBuffer, 
				tParam.tConfig.sErrMsg,
				tParam.tConfig.sErrMsg.Len(),
				PIIOBUF_NONE );
			return 0;
			}
		return (tFn)( *this, tPIHTTP, pVariable, iVariableLen, pValue,
			iValueLen, tParam );
		};

	/* ---
	SSI Parse the memory block pMap and send the result to the remote
	client
	Returns 0 on error.
	--- */
	int SSIParse( PIHTTP &tPIHTTP, PIFInfo *pFInfo, const char *pMap,
		int iLen, int iFlags )
		{
#define MIN(a,b) ((a)<(b)?(a):(b))
		SSIConfig tLocalConfig( tConfig );
		SSIParam tParam( pFInfo, tLocalConfig, pExecFileFixup );
		for(;;)
			{
			/* --- find next ssi directive --- */
			int i;
			for( i=0; i<iLen &&
				strncmp( &(pMap[i]), "<!--#", MIN(5,iLen-i) ); i++ );

			/* --- send what has been parsed so far --- */
			if ( i )
				{
				PIIOBuffer_write(
					tPIHTTP.pBuffer, 
					pMap,
					i,
					iFlags );
				};
			if ( i==iLen ) { break; };	/* done! */
			i += 5; pMap = &(pMap[i]); iLen -= i; i=0;

			/* ---
			this is an SSI directive
			--- */
			/* --- scan forward to next '-->', skipping quoted text --- */
			int iQuoted = 0;
			for( ; i<iLen &&
				( strncmp( &(pMap[i]), "-->", MIN(3,iLen-i)) || iQuoted ); i++)
				{
				if ( pMap[i]=='"' ) { iQuoted = ( iQuoted ? 0 : 1 ); };
				};

			/* --- if EOF before '-->' write out an error --- */
			if ( i==iLen )
				{
				PIIOBuffer_write(
					tPIHTTP.pBuffer, 
					tParam.tConfig.sErrMsg,
					tParam.tConfig.sErrMsg.Len(),
					iFlags );
				break;
				};

			/* --- process the directive, ignore errors --- */
			ProcessSSIDirective( tPIHTTP, pMap, i, tParam );

			/* --- skip the directive and continue --- */
			i += 3; pMap = &(pMap[i]); iLen -= i;
			};
		return 1;	
		};

	int Handle( int iPhase, PIHTTP &tPIHTTP, PIIOBuffer &/* tBuffer */ )
		{
		if ( iPhase!=PH_HANDLE )
			{ return PIAPI_ERROR; };

		PIDB *pR = tPIHTTP.pResponseDB;
		PIDB *pQ = tPIHTTP.pRequestDB;

		/* ---
		Get file object
		--- */
		const char *pPath = (const char *)PIDB_lookup( pR, PIDBTYPE_STRING,
			KEY_INT_PATH, 0 );
		PIFInfo *pFInfo = HTTPCore_getCachedFile( pPath ); 
		if ( !pFInfo )
			{ return PIAPI_ERROR; };

		/* ---	
		This file must exist and be a regular file 
		--- */
		if ( !PIFInfo_exists( pFInfo ) ||
			!PIFInfo_isRegular( pFInfo ) )
			{
			HTTPCore_logError( &tPIHTTP, "SendSSI: File with path '%s' \
does not exist or is not a regular file.", pPath );
			HTTPCore_releaseCachedFile( pFInfo );	
			return HTTPUtil_doHTTPError( &tPIHTTP, ST_INTERNALERROR );
			};

		/* ---
		memory map the file
		--- */

		/* --- get the file map --- */
		const char *pMap = 0;
		int iLen = 0;

		/* --- get filemap object --- */
		PIFMap *pFileMap = 0;
		if ( PIFInfo_getSize( pFInfo )>0 )
			{
			pFileMap = PIFMap_new( pFInfo );
			if ( pFileMap )
				{
				pMap = (const char *)PIFMap_lock( pFileMap, &iLen );
				};
			if ( !pFileMap || !pMap )
				{	
				HTTPCore_logError( &tPIHTTP, "HTTPCore_sendFile: \
Could not memory map file with path '%s'.", pPath );
				PIFMap_delete( pFileMap );
				HTTPCore_releaseCachedFile( pFInfo );	
				return HTTPUtil_doHTTPError( &tPIHTTP, ST_INTERNALERROR );
				};
			}
		else
			{
			/* --- zero length file --- */
			pMap = "";
			iLen = 0;
			};

		/* --- set content type to 'text/html --- */
		PIDB_replace( pR, PIDBTYPE_RFC822, KEY_HTTP_CONTENTTYPE,
			(void *)"text/html", 0 );

		/* --- dynamically generated content should not be cached --- */
		PIDB_replace( pR, PIDBTYPE_RFC822, KEY_HTTP_PRAGMA, (void *)"no-cache", 0 );
		PIDB_replace( pR, PIDBTYPE_RFC822, KEY_HTTP_EXPIRES,	
			(void *)"Mon, 23 Feb 1970 15:30:00 GMT", 0 );

		/* --- we don't send a content-length so disable keep open --- */
		PIDB_replace( tPIHTTP.pConnectionDB, PIDBTYPE_OPAQUE, KEY_INT_KEEPOPEN,
			(void *)0, 0 );

		/* --- last modified --- */
		Pi3String *pTmp = Pi3String_new( 0 );
		time_t tT;
		time( &tT );
		PIPlatform_beforeUnsafeBlock();
		struct tm *pTms = gmtime( &tT );
		HTTPUtil_rFC822Time( pTms, pTmp );
		PIPlatform_afterUnsafeBlock();
		PIDB_add( pR, PIDBTYPE_RFC822, KEY_HTTP_LASTMODIFIED,
			(void *)Pi3String_getPtr( pTmp ), 0 );
		Pi3String_delete( pTmp );

		/* --- If current status is 0 then change to 200 OK --- */
		if ( !tPIHTTP.iStatus )
			{
			tPIHTTP.iStatus = ST_OK;
			};

		/* ---
		Lookup method 
		--- */
		int iMethod = (int)PIDB_lookup( pQ, PIDBTYPE_OPAQUE,
			KEY_HTTP_METHOD, 0 );

		/* --- 	
		Get the header
		--- */
		enum { PATTERN_BUF_SIZE=1024 };
		char szHeader[PATTERN_BUF_SIZE];
		char *pHeader = 0;
		int iHeaderLen = 0;
		if ( pHeaderPattern )
			{
			pHeader = ( iMethod==MD_GET ) ? szHeader : 0;
			iHeaderLen = Pi3Expression_write( pHeaderPattern, &tPIHTTP, 0,
					pHeader, PATTERN_BUF_SIZE );
			if ( iHeaderLen>PATTERN_BUF_SIZE && pHeader )
				{
				pHeader = (char *)PIHTTP_allocMem( &tPIHTTP, iHeaderLen );
				Pi3Expression_write( pHeaderPattern, &tPIHTTP, 0, pHeader,
					iHeaderLen );
				};
			};

		/* --- 	
		Get the footer
		--- */
		char szFooter[PATTERN_BUF_SIZE];
		char *pFooter = 0;
		int iFooterLen = 0;
		if ( pFooterPattern )
			{
			pFooter = ( iMethod==MD_GET ) ? szFooter : 0;
			iFooterLen = Pi3Expression_write( pFooterPattern, &tPIHTTP, 0,
					pFooter, PATTERN_BUF_SIZE );
			if ( iFooterLen>PATTERN_BUF_SIZE && pFooter )
				{
				pFooter = (char *)PIHTTP_allocMem( &tPIHTTP, iFooterLen );
				Pi3Expression_write( pFooterPattern, &tPIHTTP, 0, pFooter,
					iFooterLen );
				};
			};

		/* --- transfer-encoding --- */
		int iFlags = (!PIHTTP_isChild( &tPIHTTP) && (int)PIDB_lookup( tPIHTTP.pRequestDB,
			PIDBTYPE_OPAQUE, KEY_HTTP_PROTOCOL, 0 ) == 3 ) ? PIIOBUF_CHUNKED : PIIOBUF_NONE;
		if ( iFlags & PIIOBUF_CHUNKED )
			{
			PIDB_replace( pR, PIDBTYPE_RFC822, KEY_HTTP_TRANSFERENCODING,
				(void *)KEY_HTTP_CHUNKED, 0 );
			}

		/* --- start response --- */
		if (	HTTPCore_sendGeneralHeaders( &tPIHTTP ) ||
				HTTPCore_sendEntityHeaders( &tPIHTTP, pR ) )
			{
			HTTPCore_releaseCachedFile( pFInfo );	
			return PIAPI_ERROR;
			};

		if ( iFlags & PIIOBUF_CHUNKED ) PIIOBuffer_flush( tPIHTTP.pBuffer );

		/* --- 
		If the Method is 'GET' return PIAPI_CONTINUE to let the 
		handling continue by sending the file.
		Otherwise assume Method is 'HEAD' and return PIAPI_COMPLETED as
		end of transaction
		--- */
		int iRet = 1;
		switch( iMethod )
			{
			case MD_GET:
				if ( pHeader && iHeaderLen )
					{
					iRet = ( PIIOBuffer_write( tPIHTTP.pBuffer, pHeader,
						iHeaderLen, iFlags)>0 );
					};
				if ( iRet )
					{
					/* --- process the file, ignore errors --- */
					iRet = SSIParse( tPIHTTP, pFInfo, pMap, iLen, iFlags );
					};
				if ( iRet && pFooter && iFooterLen )
					{
					iRet = ( PIIOBuffer_write( tPIHTTP.pBuffer, pFooter,
						iFooterLen, iFlags)>0 );
					};
				break;
		
			case MD_HEAD:
				break;

			default:
				assert( 0 );
				if ( pFileMap ) { PIFMap_delete( pFileMap ); };
				HTTPCore_releaseCachedFile( pFInfo );	
				return PIAPI_ERROR;
			};

		if ( pFileMap )
			{ PIFMap_delete( pFileMap ); };
		HTTPCore_releaseCachedFile( pFInfo );	

		/* ---
		If iRet is 0 an error occurred
		** NOTE:
		--- */

		return PIAPI_COMPLETED;
		};
};

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int SendSSI_constructor( PIObject *pObj,
	int iArgc, const char *ppArgv[] )
{
	return HandlerBaseHTTP_constructor( pObj, PI_NEW( SendSSI( pObj,
		iArgc, ppArgv ) ) );
}

#if 0
/*___+++CNF_BEGIN+++___*/
	<Class>
		Name SendSSIClass
		Type LogicExtension
		Library HTTP
		OnClassLoad HandlerBaseHTTP_onClassLoad
		Constructor SendSSI_constructor
		CopyConstructor HandlerBaseHTTP_copyConstructor
		Destructor HandlerBaseHTTP_destructor
		Execute HandlerBaseHTTP_execute
	</Class>

	<Object>
		Name SendSSI
		Class SendSSIClass
	</Object>
/*___+++CNF_END+++___*/
#endif


