/*____________________________________________________________________________*\
 *

 Copyright (c) 1997-2003 John Roy, Holger Zimmermann. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. The name of the author may not be used to endorse or promote products
    derived from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/Server/MMThread.cpp,v $
 * $Date: 2003/06/09 14:44:32 $
 *
 Description:
	IO server class, derived from IOServerBase class.

	Multithreaded dispatch of IO requests.

\*____________________________________________________________________________*/
//$SourceTop:$

#include <assert.h>
#include <iostream.h>
#include <stdlib.h>
#include <string.h>

#include "IOSrvBse.h"
#include "Pi2API.h"
#include "PIStrStr.h"
#include "DblList.h"

/*____________________________________________________________________________*\
 *
 Description:
\*____________________________________________________________________________*/
#define KEY_CONF_MAXTHREADS		"MaxThreads"
#define KEY_CONF_EXITAFTER		"ExitAfter"
#define KEY_CONF_SECONDARYIO	"SecondaryIO"

#define CONFIG_ERR(this, msg)   \
	{ PILog_addMessage( PIObject_getDB(this), \
	PIObject_getConfigurationDB(this), PILOG_ERROR, (msg) ); }
#define CONFIG_WARN(this, msg)   \
	{ PILog_addMessage( PIObject_getDB(this), \
	PIObject_getConfigurationDB(this), PILOG_WARNING, (msg) ); }
/*
#define D { cerr << __FILE__ << ": " << __LINE__ << endl; };
*/
#define D

/* ---
	MUSTBETRUE - alway execute the statement, additionally assert it
	during development 
--- */
#if !defined(NDEBUG)
#	define MUSTBETRUE(x)	(assert(x))
#else
#	define MUSTBETRUE(x)	(x)
#endif

/* --
	Determine whether we use thread-per-request or thread-reuse model
--- */
#ifdef WIN32
#	define THREAD_PER_REQUEST 0
#else
#	define THREAD_PER_REQUEST 1	/* thread-reuse not ready for prime time! */
#endif

/*____________________________________________________________________________*\
 *
 Description:
	Documentation.
\*____________________________________________________________________________*/
#if 0
	/*
	** HTML documentation for this server
	*/
/*___+++HTMLDOC_BEGIN+++___*/
Name:
	MTMIOServer

Description:
	This object performs multithreaded acceptance and dispatch of IO requests.
      On execution it waits for a new IO request on multiple interfaces and dispatches
      it to a worker thread for processing. It then returns to dispatch another request
      and the worker thread is reused after processing the request.
	<P>
	This dispatching object initializes a primary and optional multiple secondary IO
      objects for accepting IO requests, and a logic object for handling those requests.

Options:
<H5>Overview</H5>
<TABLE BORDER=1>
<TH>Option
<TH>Default
<TH>Values
<TH>Short Description
<TH>Example(s)

<TR>
<TD>IOObject
<TD>+
<TD>A Pi3 IO object name
<TD>A Pi3 object
<TD>IOObject "ServerTCPIPObject"

<TR>
<TD>SecondaryIO
<TD>-
<TD>A Pi3 IO object name
<TD>A Pi3 object
<TD>SecondaryIO "SSLIOObject"

<TR>
<TD>LogicObject
<TD>+
<TD>A Pi3 Logic object name
<TD>A Pi3 object
<TD>LogicObject "HandleRequestObject"

<TR>
<TD>MaxThreads
<TD>+
<TD>A non-negative
<TD>Number of threads to use
<TD>MaxThreads 10

<TR>
<TD>ExitAfter
<TD>-1
<TD>An integer
<TD>Number of request to exit after
<TD>ExitAfter 1000

</TABLE>
<STRONG>-</STRONG> in the <IT>default</IT> indicates no default<BR>
<STRONG>+</STRONG> in the <IT>default</IT> indicates the field is mandatory<BR>

<H4>Description of Options</H4>
<H5>
	IOObject
</H5>
Specifies the primary IO object used to accept new IO requests. The IO object
is created during construction. During execution of this object PIObject_copy()
is invoked to create a copy IO object for each new request. When processing
of the request is completed, the copied IO object is destroyed and the server
waits for another IO request by again calling PIObject_copy() on the main
IO object.
The IO object can be an object which provides an IO service (such as TCP/IP,
named Pipe or shared memory IO) or an IO filter which performs translation
on such an IO object (such as encryption or monitoring).

<H5>
	SecondaryIO
</H5>
Specifies a secondary IO object used to accept new IO requests. This option can
occur multiple times within the configuration. An accept thread is created for
each IO object in order to accept concurrent connections from different
interfaces. The behaviour of the IO object is the same as described for option
IOObject.

<H5>
	LogicObject
</H5>
Specifies the logic object used to handle IO requests. The object is 
created during construction and its <I>execute</I> method invoked
(using PILogic_execute()) for each IO request accepted. The IO object
for the request is passed as an argument to this logic object.

<H5>
	MaxThreads
</H5>
Specifies the maximum number of threads to use to concurrently handle
IO requests. While no more threads are availablable to new IO requests
will be accepted.
This number should be between 1 and 1000.

<H5>
	ExitAfter
</H5>
Specifying a positive number for this value will cause the this server
object to call exit() after that number of requests. Threads with outstanding
requests will be allow to complete before exit() is invoked.
<P>
This should only be used in the context of a process group where a monitor
process fork()s child processes to replace ones that die. In that
scenario it is good for preventing longterm memory or resource leaks.
<P>
This directive has no effect on non-UNIX systems.

Returns:
	PIAPI_COMPLETED on success.
	PIAPI_ERROR and PIAPI_ABORT respectively for generic and severe
	error conditions.

Example:
	<PRE>
	&lt;Object&gt;
		Name MTMIOServer
		Class MTMIOServerClass
		IOObject ServerTCPIPIO
		SecondaryIO SSLIOObject
		SecondaryIO ServerTCPIPIO BindHost="localhost" BindPort="8080"
		LogicObject HandleHTTP
	&lt;/Object&gt;
	</PRE>
/*___+++HTMLDOC_END+++___*/
#endif

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class MTMIOServer : public IOServerBase
{
private:
	class ThreadRecord
		{
		public:
		PIThread *pThread;
#if !THREAD_PER_REQUEST
		PISync *pSema;
#endif
		MTMIOServer *pServer;
		PIObject *pIOObject;
		int iRet;
		ThreadRecord( PIThread *pTheThread )
			:	pThread( pTheThread ), iRet( PIAPI_COMPLETED ),
#if !THREAD_PER_REQUEST
				pSema( PIPlatform_allocLocalSemaphore( 0, 2 ) ),
#endif
				pServer( 0 ),
				pIOObject( 0 )
			{
			};

		~ThreadRecord()
			{
#if !THREAD_PER_REQUEST
			PISync_delete( pSema );
#endif
			};
		};
	PISync *pFreeThreadsMutex;		/* guard access to free threads array */
	PISync *pFreeThreadsSemaphore;	/* # threads is a resource, count usage */
	int iNumBusyThreads;			/* count the number of busy threads */
	ThreadRecord **ppAllThreads;	/* array of all thread records */
	ThreadRecord **ppFreeThreads;	/* array of records of free threads */
	int iMaxThreads;
	int iRequestsToLive;			/* exit after this number of requests */
	ThreadRecord **ppAcceptThreads;
	int iAcceptThreads;
	DblList lAccept;				/* patterns for secondary IO objects */
	PISync *pAcceptThreadsSemaphore;	/* # accept threads is a resource */

	/* ---
	Automatically invoked from IOServerBase::Accept()
	This overrides the default simple DoDispatchConnection() in
	IOServerBase (which just calls DispatchConnection() directly, to
	allocate a thread and make the thread run DispatchConnection()
	--- */
	virtual int DoDispatchConnection( PIObject *pIOObject )
		{
		/* ---
		Select a thread to run, access to the free threads list is
		protected by the pFreeThreadsMutex.
		--- */
		MUSTBETRUE( !PISync_lock( pFreeThreadsMutex ) );
		ThreadRecord *pThreadRecord = 0;
		for( int i=0; i<iMaxThreads; i++ )
			{
			if ( ppFreeThreads[i] )
				{
				pThreadRecord = ppFreeThreads[i];
				ppFreeThreads[i] = 0;
				break;
				};
			};
		/* ---	
		Another thread is officially busy
		--- */
		iNumBusyThreads++;
		MUSTBETRUE( !PISync_unlock( pFreeThreadsMutex ) );
		PIThread *pThread = pThreadRecord->pThread;
		assert( pThreadRecord->pThread );
		assert( pThread );
		
#if THREAD_PER_REQUEST
		/* ---
		Wait for this thread to fully terminate, this prevents
		a race condition where a thread adds itself to the free
		list but gets preempted before terminating
		--- */
		PIThread_waitForThread( pThread );
		pThreadRecord->pIOObject = pIOObject;
		pThreadRecord->pServer = this;
		if ( PIThread_begin(
			pThread,
			(PIThreadFn)MTMIOServer::StaticDispatchConnection,
			(unsigned long)pThreadRecord,
			PITHREAD_PRIORITY_HIGH, 0 ) ) 
			{
			MUSTBETRUE( !PISync_unlock( pFreeThreadsSemaphore ) );
			MUSTBETRUE( !PIObject_delete( pIOObject, 0, 0 ) );
	
			/* --- consider this a critical error --- */
			return PIAPI_ABORT;
			};
#else
		pThreadRecord->pIOObject = pIOObject;
		MUSTBETRUE( !PISync_unlock( pThreadRecord->pSema ) );
#endif
		return PIAPI_COMPLETED;
		};

	/* ---
	Called by StaticDispatchConnection() this function	
	marks the current thread as free so it may be reused
	--- */
	void FreeThread( ThreadRecord *pFree )
		{

		/* --- add this thread back to the free list --- */
		assert( pFreeThreadsMutex );
		MUSTBETRUE( !PISync_lock( pFreeThreadsMutex ) );
		int i=0; for(; i<iMaxThreads; i++ )
			{
			if ( !ppFreeThreads[i] )
				{ ppFreeThreads[i] = pFree; break; };
			};
		/* ---
		Another thread is officially free
		--- */
		iNumBusyThreads--;
		MUSTBETRUE( !PISync_unlock( pFreeThreadsMutex ) );
		MUSTBETRUE( !PISync_unlock( pFreeThreadsSemaphore ) );
		assert( i<iMaxThreads );
		};

	/* ---
	Called by the threads begin function this runs in a thread
	created for this connection.
	--- */
	inline static int StaticDispatchConnection( ThreadRecord *pThreadRecord )
		{
		assert( pThreadRecord );
		if ( !pThreadRecord )
			{ return PIAPI_ERROR; }

		for(;;)
			{
			/* --- wait on the mutex --- */
#if !THREAD_PER_REQUEST
			if ( PISync_lock( pThreadRecord->pSema ) )
				{
				MUSTBETRUE( 0 );
				};
#endif

			MTMIOServer *pServer = pThreadRecord->pServer;
			if ( !pServer )
				{
				/* --- message to exit the thread --- */
				break;
				};

			/* ---
			Before attempting to dispatch a connection, make sure there
			is a thread available to handle it
			--- */
			MUSTBETRUE( !PISync_lock( pServer->pFreeThreadsSemaphore ) );

			PIObject *pIOObject = pThreadRecord->pIOObject;
 			pServer->DispatchConnection( pIOObject );
			pServer->FreeThread( pThreadRecord );

#if THREAD_PER_REQUEST
			break;
#endif
			};
		return PIAPI_COMPLETED;
		}

	/* ---
	Called by the threads begin function this runs in a thread
	created for this connection.
	--- */
	inline static int StaticAcceptConnection( ThreadRecord *pThreadRecord )
		{
		assert( pThreadRecord );
		if ( !pThreadRecord )
			{ return PIAPI_ERROR; }

		for(;;)
			{
			/* --- wait on the mutex --- */
#if !THREAD_PER_REQUEST
			if ( PISync_lock( pThreadRecord->pSema ) )
				{
				MUSTBETRUE( 0 );
				};
#endif

			MTMIOServer *pServer = pThreadRecord->pServer;
			if ( !pServer )
				{
				/* --- message to exit the thread --- */
				break;
				};


			/* ---
			Accept new IO connection
			--- */
			PIObject *pIOObject = PIObject_copy( pThreadRecord->pIOObject, 0, 0 );

			if ( !pIOObject )
				{
				/* --- the program is terminating --- */
				break;
				};

			/* ---
			Dispatch to worker thread
			--- */
			pThreadRecord->iRet = pServer->DoDispatchConnection( pIOObject );

			/* ---
			Unlock main accept loop
			--- */
#if THREAD_PER_REQUEST
			pThreadRecord->pServer = 0;
#endif
			MUSTBETRUE( !PISync_unlock( pServer->pAcceptThreadsSemaphore ) );
			};
		return PIAPI_COMPLETED;
		}

	static int ParameterFn( void *pData, const char *pVar, const char *pVal,
		const char *pWhere )
		{
		assert( pData );
		return ((MTMIOServer *)pData)->Parameter( pVar, pVal,
			pWhere ? pWhere : "" );
		};

protected:
	int Parameter( const char *pVariable, const char *pValue,
		const char *pWhere )
		{
		assert( pVariable && pValue );
		PIOStrStream os;
		os << pWhere << "MTMIOServer: ";

		if ( !PIUtil_stricmp( KEY_CONF_EXITAFTER, pVariable ) )
			{
			iRequestsToLive = atoi( pValue );
			iRequestsToLive = iRequestsToLive > 0 ? iRequestsToLive : -1 ;
			}

		else if ( !PIUtil_stricmp( KEY_CONF_MAXTHREADS, pVariable ) )
			{
			iMaxThreads = atoi( pValue );
			if ( iMaxThreads<1 || iMaxThreads>1000 )
				{
				CONFIG_WARN( pObject, "MTMIOServer: Invalid number for \
'MaxThreads', defaulting to 25");
				};
			}
		else if ( !PIUtil_stricmp( KEY_CONF_SECONDARYIO, pVariable ) )
			{
			lAccept.Append( (DblList::type) PI_NEW( PIString( pValue ) ) );
			iAcceptThreads++;
			}
		else if ( !PIUtil_stricmp( KEY_CONF_IOOBJECT, pVariable ) )
			{
			// has already been loaded by parent class
			iAcceptThreads++;
			}
		else if ( !PIUtil_stricmp( KEY_CONF_LOGICOBJECT, pVariable ) )
			{
			// has already been loaded by parent class
			}
		else
			{
			os << "Unknown parameter '" << pVariable << "'" << ends;
			CONFIG_ERR( pObject, os.str() );
			return 0;
			};

		return 1;
		};

public:
	MTMIOServer( const PIObject *pTheObject, int iArgc, const char *pArgv[] )
	:	IOServerBase( pTheObject ),
		pFreeThreadsMutex( PIPlatform_allocLocalMutex() ),
		pFreeThreadsSemaphore( 0 ),
		pAcceptThreadsSemaphore( PIPlatform_allocLocalSemaphore( 0, 2 )),
		iNumBusyThreads( 0 ),
		ppAllThreads( 0 ),
		ppFreeThreads( 0 ),
		ppAcceptThreads( 0 ),
		iMaxThreads( 25 ),
		iAcceptThreads( 0 ),
		iRequestsToLive( -1 )
		{
		if ( !bOK ) { return;	/* Base class initialization failed */ };

		bOK = PIObject_readParameters( (PIObject *)pObject, iArgc, pArgv, ParameterFn, this );
		if ( !bOK ) { return;	/* parameter function failed */ };

		bOK = 0;				/* until initialization is complete */
		if ( !pFreeThreadsMutex )
			{
			CONFIG_ERR( pObject, "MTMIOServer: Could not allocate mutex.");
			return;
			};

		pFreeThreadsSemaphore = PIPlatform_allocLocalSemaphore( iMaxThreads,
			iMaxThreads );
		if ( !pFreeThreadsSemaphore )
			{
			CONFIG_ERR( pObject, "MTMIOServer: Could not allocate semaphore.");
			return;
			};
		ppAllThreads = new ThreadRecord *[iMaxThreads];
		ppFreeThreads = new ThreadRecord *[iMaxThreads];
		ppAcceptThreads = new ThreadRecord *[iAcceptThreads];
		if ( !ppFreeThreads || !ppAllThreads || !ppAcceptThreads )
			{ assert( 0 ); return; };
		memset( ppAllThreads, 0, sizeof( ThreadRecord * ) * iMaxThreads );
		memset( ppFreeThreads, 0, sizeof( ThreadRecord * ) * iMaxThreads );
		memset( ppAcceptThreads, 0, sizeof( ThreadRecord * ) * iAcceptThreads );

		int iGood = 1;

		/* ---  begin accept thread for main I/O object --- */
		PIThread *pNew = PIThread_new( 0, 0 );
		ThreadRecord *pRecord = new ThreadRecord( pNew );
		pRecord->pIOObject = GetPrototypeIOObject();
#if THREAD_PER_REQUEST
		pRecord->pServer = 0;
#endif

#if !THREAD_PER_REQUEST
		pRecord->pServer = this;
#endif

		int i = 0;
		ppAcceptThreads[i++] = pRecord;
		if ( PIThread_begin(
			pNew,
			(PIThreadFn)MTMIOServer::StaticAcceptConnection,
			(unsigned long)pRecord,
			PITHREAD_PRIORITY_HIGH, 0 ) ) 
			{
				PIThread_delete( pNew );
				/* --- consider this a critical error --- */
				iGood = 0;
				assert( 0 );
				CONFIG_ERR( pObject,
"MTMIOServer: Failed to prime thread, PIThread_begin() failed.");
			};

		/* ---  begin accept threads for secondary I/O objects --- */
		for( DblListIterator a( lAccept ); !a.BadIndex(); a++)
			{
			if ( iAcceptThreads == i ) break;
			PIThread *pNew = PIThread_new( 0, 0 );
			ThreadRecord *pRecord = new ThreadRecord( pNew );
			pRecord->pIOObject = PIObject_loadFromLine( PIObject_getDB(pObject),
				PIObject_getConfigurationDB(pObject), (const char *)a.Current() );
			if ( !pRecord->pIOObject )
				{
				iGood = 0;
				pRecord->pServer = 0;
				CONFIG_ERR( pObject,
"MTMIOServer: Failed to load IOObject, PIObject_loadFromLine() failed." );
				break;
				}

#if THREAD_PER_REQUEST
			pRecord->pServer = 0;
#endif

#if !THREAD_PER_REQUEST
			pRecord->pServer = this;
#endif
			ppAcceptThreads[i++] = pRecord;
			if ( PIThread_begin(
				pNew,
				(PIThreadFn)MTMIOServer::StaticAcceptConnection,
				(unsigned long)pRecord,
				PITHREAD_PRIORITY_HIGH, 0 ) ) 
				{
					PIThread_delete( pNew );
					/* --- consider this a critical error --- */
					iGood = 0;
					assert( 0 );
					CONFIG_ERR( pObject,
"MTMIOServer: Failed to prime thread, PIThread_begin() failed.");
				};
			}

		for( i=0; iGood && i<iMaxThreads; i++ )
			{
			PIThread *pNew = PIThread_new( 0, 0 );
			ThreadRecord *pRecord = new ThreadRecord( pNew );
			ppFreeThreads[i] = pRecord;
			ppAllThreads[i] = pRecord;
#if !THREAD_PER_REQUEST
			pRecord->pServer = this;
			if ( PIThread_begin(
				pNew,
				(PIThreadFn)MTMIOServer::StaticDispatchConnection,
				(unsigned long)pRecord,
				PITHREAD_PRIORITY_MED, 0 ) ) 
				{
				assert( 0 );
				CONFIG_ERR( pObject,
"MTMIOServer: Failed to prime thread, PIThread_begin() failed.");
				iGood = 0;
				};
#endif
			};

		if ( ppAllThreads && ppFreeThreads && iGood )
			{ bOK = 1; };
		};

	virtual ~MTMIOServer()
		{
		for( DblListIterator a( lAccept ); !a.BadIndex(); a++)
			{ PI_DELETE( (PIString *)a.Current() ); };

		if ( !ppFreeThreads || !ppAllThreads || !ppAcceptThreads )
			{ assert( 0 ); return; };

		int i;
		for( i=0; i<iAcceptThreads; i++ )
			{
			if ( !ppAcceptThreads[i] ) continue;
			PIThread *pThread = ppAcceptThreads[i]->pThread;
#if !THREAD_PER_REQUEST
			/* --- signal thread to die --- */
			ppAcceptThreads[i]->pServer = 0;
//				MUSTBETRUE( !PIThread_terminate( pThread, 0 ) );
			if ( PIObject_delete( ppAcceptThreads[i]->pIOObject, 0, 0 ) == PIAPI_COMPLETED
				&& pIO && ppAcceptThreads[i]->pIOObject == pIO )
				{
				pIO = 0;
				}
			PISync_unlock( ppAcceptThreads[i]->pSema );
#endif
			MUSTBETRUE( !PIThread_waitForThread( pThread ) );
			MUSTBETRUE( !PIThread_delete( pThread ) );
			delete ppAcceptThreads[i];
			}

		for( i=0; i<iMaxThreads; i++ )
			{
			if ( !ppAllThreads[i] ) continue;
			PIThread *pThread = ppAllThreads[i]->pThread;
#if !THREAD_PER_REQUEST
			/* --- signal thread to die --- */
			ppAllThreads[i]->pServer = 0;
			PISync_unlock( ppAllThreads[i]->pSema );
			MUSTBETRUE( !PIThread_waitForThread( pThread ) );
#endif
			MUSTBETRUE( !PIThread_delete( pThread ) );
			delete ppAllThreads[i];
			}

		delete [] ppAllThreads;
		delete [] ppFreeThreads;
		delete [] ppAcceptThreads;

		MUSTBETRUE( 
			!pFreeThreadsMutex || !PISync_delete( pFreeThreadsMutex ) );
		MUSTBETRUE(
			!pFreeThreadsSemaphore || !PISync_delete( pFreeThreadsSemaphore ) );
		MUSTBETRUE(
			!pAcceptThreadsSemaphore || !PISync_delete( pAcceptThreadsSemaphore ) );
		};

	virtual int Accept()
		{
		/* ---
		Check is it time for this server to exit
		--- */
#if POSIX
		if ( iRequestsToLive )
			{ 	
			if ( iRequestsToLive>0 )
				{ iRequestsToLive--; };
			}
		else
			{
			/* ---
			Time to go, gracefully exit.
			--- */
			while( iNumBusyThreads>0 )
				{
				MUSTBETRUE( !PISync_lock( pFreeThreadsSemaphore ) );
				};
#if 1
			/*
			** Why even bother?
			*/
			/* ---
			Allow all threads to die
			--- */
			for( int i=0; i<iMaxThreads; i++ )
				{
				/* --- signal thread to die --- */
				ppAllThreads[i]->pServer = 0;

#if !THREAD_PER_REQUEST
				PISync_unlock( ppAllThreads[i]->pSema );
				MUSTBETRUE( !PIThread_terminate( pThread, 0 ) );
#endif		
				};
#endif
			::exit( 0 );
			};
#endif

		// run the accept threads
		int iRet = PIAPI_COMPLETED;
#if !THREAD_PER_REQUEST
		for( int i=0; i<iAcceptThreads; i++ )
			{
			if ( PISync_tryLock(ppAcceptThreads[i]->pSema)==PIAPI_WOULDBLOCK )
				{
				int iRet = ppAcceptThreads[i]->iRet;
				if ( iRet == PIAPI_ABORT ) break;

				/* --- signal thread to run --- */
				PISync_unlock( ppAcceptThreads[i]->pSema );
				}
			}
#endif

#if THREAD_PER_REQUEST
		for( int i=0; i<iAcceptThreads; i++ )
			{
			if ( ppAcceptThreads[i]->pServer == 0 )
				{
				int iRet = ppAcceptThreads[i]->iRet;
				if ( iRet == PIAPI_ABORT ) break;

				/* --- signal thread to run --- */
				ppAcceptThreads[i]->pServer = this;
				PIThread_waitForThread( ppAcceptThreads[i]->pThread );
				if ( PIThread_begin(
				ppAcceptThreads[i]->pThread,
				(PIThreadFn)MTMIOServer::StaticAcceptConnection,
				(unsigned long)ppAcceptThreads[i],
				PITHREAD_PRIORITY_HIGH, 0 ) ) 
					{
					iRet = PIAPI_ABORT;
					};
				}
			}
#endif

		/* ---
		Block until one accept thread did its work
		--- */
		do {}
		while ( PISync_tryLock(pAcceptThreadsSemaphore)==PIAPI_WOULDBLOCK );
//		MUSTBETRUE( !PISync_lock( pAcceptThreadsSemaphore ) );
		return iRet;
		};
};

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int MTMIOServer_onClassLoad( PIClass_LoadAction eLoad, void * )
{
	switch( eLoad )
		{
		case STARTUP:
		default:;
		};

	return PIAPI_COMPLETED;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int MTMIOServer_constructor( PIObject *pObj, int iArgc, const char *pArgv[] )
{
	MTMIOServer *pServer = new MTMIOServer( pObj, iArgc, pArgv );
	if (!( pServer && pServer->IsOK() ))
		{
		delete pServer;
		return PIAPI_ERROR;
		};
	if ( PIObject_setUserData( pObj, pServer ) )
		{
		CONFIG_ERR( pObj, "MTMIOServer_constructor: PIObject_setUserData() failed" );
		return PIAPI_ERROR;
		};
	return PIAPI_COMPLETED;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int MTMIOServer_copyConstructor( PIObject *, int,
	const char *[] )
{
	/* --- forbid copy construction --- */
	return PIAPI_ERROR;
}

#if 0
/*___+++CNF_BEGIN+++___*/
	<Class>
		Name MTMIOServerClass
		Type LogicExtension
		Library Server
		OnClassLoad MTMIOServer_onClassLoad
		Constructor MTMIOServer_constructor
		CopyConstructor MTMIOServer_copyConstructor
		Destructor IOServerBase_destructor
		Execute IOServerBase_execute
	</Class>

	<Object>
		Name MTMIOServer
		Class MTMIOServerClass
	</Object>

/*___+++CNF_END+++___*/
#endif

